/*
 *  Copyright (c) 2004 Forward Computing and Control Pty. Ltd.
 *  NSW. Australia,  www.forward.com.au
 *  All rights reserved.
 *  Redistribution of the source of this package, with or without
 *  modification, is NOT permitted.
 *  Redistribution of the compiled java .class files of this package, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions in compiled java .class form must reproduce the above
 *  copyright notice, this list of conditions and the following disclaimer
 *  in the documentation and/or other materials provided with the distribution.
 *  2. If the software is modified, the compiled java .class distribution must
 *  include, in the documentation and/or other materials provided with the
 *  distribution, a statement that the software has been modified and
 *  identify the person who was responsable for the modifications.
 *  3. Neither the name of the author nor Forward Computing and Control may be
 *  used to endorse or promote products derived from this software
 *  without specific prior written permission.
 *  4.  All advertising materials mentioning features or use of this
 *  software must display the following acknowledgment:
 *  "This product includes software developed by Matthew Ford
 *  and Forward Computing and Control Pty. Ltd.
 *  (http://www.forward.com.au)."
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
 
/* change log
 *  16th October 2004
 *  Original version.
 *
 *  20th October 2004 
 *  V1.2.0 added KeyEventDispatcher
 *  remove tip on next key press
 *  still use focus listener to remove on mouse focus change
 *
 *  5th January 2006
 *  V1.3.0 added support for input tips on JTable cells and optional attachment of focus and keyboard listener.
 */
 
package au.com.forward.inputTipsManager;

import java.awt.*;

import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import javax.swing.JApplet;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JTable;
import javax.swing.JToolTip;
import javax.swing.JViewport;
import javax.swing.JWindow;
import javax.swing.Popup;
import javax.swing.PopupFactory;
import javax.swing.SwingUtilities;


/**
 *  Manages all the <code>InputTips</code> in the application. <p>
 *
 *  use <br>
 *  <code>InputTipsManager.sharedInstance().showInputTip(jcomponent,tip);</code>
 *  <br>
 *  to show the tip and<br>
 *  <code>InputTipsManager.sharedInstance().hideInputTip();</code><br>
 *  to hide it.<p>
 *
 *  The InputTip will also be removed when the <code>jcomponent</code> looses
 *  focus, or on next keystroke.
 *  <p>
 *  For more control use <br>
 *  <code>InputTipsManager.sharedInstance().showInputTip(jcomponent,tipText,attachFocusListener, cellRect); </code><br>
 *  which lets you control if a focus and keyboard listener is attached and specify
 * the cell rectangle in a JTable to show the tip next to.<p>
 * For example <br>
 * <code>Rectangle cellRect = table.getCellRect(selectedRow,selectedColumn, true);<br>
				String tip = "<HTML>" + "Row "+selectedRow+" has been selected.";<br>
        InputTipsManager.sharedInstance().showInputTip(table,tip, false, cellRect);</code><br>
 *
 *@author     Matthew Ford
 *@version    1.3
 *created    October 20, 2004
 */
public class InputTipsManager implements FocusListener {
	/**
	 *  Description of the Field
	 */
	private String toolTipText;

	/**
	 *  Description of the Field
	 */
	private JComponent insideComponent;

	/**
	 *  Description of the Field
	 */
	private InputTipsKeyHandler inputTipsKeyHandler = null;

	/**
	 *  Description of the Field
	 */
	private KeyboardFocusManager keyboardFocusManager = null;

	/**
	 *  Description of the Field
	 */
	final static InputTipsManager sharedInstance = new InputTipsManager();

	/**
	 *  Description of the Field
	 */
	transient Popup tipWindow;

	/**
	 *  Description of the Field
	 */
	private JToolTip tip;

	/**
	 *  Description of the Field
	 */
	private Rectangle popupRect = null;

	/**
	 *  Description of the Field
	 */
	private Rectangle popupFrameRect = null;

	/**
	 *  Description of the Field
	 */
	private boolean enabled = true;

	/**
	 *  Description of the Field
	 */
	private boolean beepEnabled = true;

	/**
	 *  Description of the Field
	 */
	private boolean tipShowing = false;

	/**
	 *  Description of the Field
	 */
	private JComponent focusComponent = null;


	/**
	 *  To satisfy FocusListener Interface does nothing
	 *
	 *@param  e  focus event
	 */
	public void focusGained(FocusEvent e) {
		// do nothing on focus gained
	}


	/**
	 *  Hide InputTip when focus lost
	 *
	 *@param  e  focus event
	 */
	public void focusLost(FocusEvent e) {
		// hide tip
		hideInputTip();
	}


	/**
	 *  Constructor for the InputTipsManager object
	 */
	protected InputTipsManager() { }


	/**
	 *  Globally enables or disables the Beep on showInputTips (default enabled).
	 *  When set to true <code>Toolkit.getDefaultToolkit().beep();</code> will be
	 *  called each time an InputTip is displayed.
	 *
	 *@param  flag  true to enable beep on input tip display, false otherwise
	 */
	public void setBeepEnabled(boolean flag) {
		beepEnabled = flag;
	}


	/**
	 *  Returns true if beeps are enabled.
	 *
	 *@return    true if <code>InputTipManger</code> will beep when displaying an
	 *      tip
	 */
	public boolean isBeepEnabled() {
		return beepEnabled;
	}


	/**
	 *  Globally enables or disables the InputTips. When set to false any existing
	 *  input tip is hidden and showInputTip() calls are ignored.
	 *
	 *@param  flag  true to enable the inputtip, false otherwise
	 */
	public void setEnabled(boolean flag) {
		enabled = flag;
		if (!flag) {
			hideInputTip();
		}
	}


	/**
	 *  Returns true if InputTips are enabled.
	 *
	 *@return    true if showInputTip() will display an tip
	 */
	public boolean isEnabled() {
		return enabled;
	}


	/**
	 *  Shows an InputTip for this component. If attachFocusListener is true, tip is removed when component looses
	 *  focus or a key is pressed or when the hideInputTip() is called.<br>
   *  If the cellRect is non-null it is used to locate the tip within the component.<br>
   *  If the component is a JTable, it is scrolled to make the cell rectangle visible.<br>
	 *  This method runs an internal method on the SwingThread using invokeLater
	 *
	 *@param  component  The JComponent to display the tip next to
	 *@param  tipText    The tip to display, can be &lt;HTML&gt; text
   *@param  attachFocusListener  true if should attach focus listener and key listener to hide tip on change in focus
   *                             false if listeners should not be attached
   *@param  cellRect   The rectangle representing the information the tip is to be displayed against. 
   *                   If the component is a JTable scroll to make this rectangle visible.
   *                   Null if not needed.
	 */
	public void showInputTip(final JComponent component, final String tipText, final boolean attachFocusListener, final Rectangle cellRect ) {
		SwingUtilities.invokeLater(
			new Runnable() {
				public void run() {
					showInputTip1(component, tipText, attachFocusListener, cellRect);
				}
			});
	}
  
	/**
	 *  Shows an InputTip for this component. Tip is removed when component looses
	 *  focus or a key is pressed or when the hideInputTip() is called.<br>
	 *  This method runs an internal method on the SwingThread using invokeLater
	 *
	 *@param  component  The JComponent to display the tip next to
	 *@param  tipText    The tip to display, can be &lt;HTML&gt; text
	 */
	public void showInputTip(final JComponent component, final String tipText) {
		showInputTip(component, tipText, true, null);
	}
  


	/**
	 *  Shows an InputTip for the component which currently has focus.<br>
	 *  If you know which component has focus use {@link
	 *  #showInputTip(JComponent,String)} instead<br>
	 *  Tip is removed when component looses focus or a key is pressed or when the hideInputTip() is
	 *  called.<br>
	 *  This method runs an internal method on the SwingThread using invokeLater
	 *
	 *@param  tipText  The tip to display, can be &lt;HTML&gt; text
	 */
	public void showInputTip(String tipText) {
		showInputTip(null, tipText,  true, null);
	}


	/**
	 *  Shows an InputTip for this component. If attachFocusListener is true, tip is removed when component looses
	 *  focus or a key is pressed or when the hideInputTip() is called.<br>
	 *
	 *@param  component  The JComponent to display the tip next to or in
	 *@param  tipText    The tip to display, can be &lt;HTML&gt; text
   *@param  attachFocusListener  true if should attach focus listener and key listener to hide tip on change in focus
   *                             false if listeners should not be attached
   *@param  cellRect   The rectangle representing the information the tip is to be displayed against. 
   *                   If the component is a JTable scroll to make this rectangle visible.
   *                   Null if not needed.
   *
	 */
	private synchronized void showInputTip1(JComponent component, String tipText, boolean attachFocusListener, Rectangle cellRect ) {
		if (component == null) {
			Component c = DefaultKeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner();
			if (c instanceof JComponent) {
				component = (JComponent) c;
			}
		}
		insideComponent = component;
		if (insideComponent == null || !insideComponent.isShowing()) {
			return;
		}

		if (enabled) {
			Dimension size;
      JViewport viewport = null;
			Point screenLocation = insideComponent.getLocationOnScreen();
			Dimension insideComponentSize = insideComponent.getSize();
      if (cellRect != null) {
        // if component is JTable scroll to the rectangle
        if (component instanceof JTable) {
          ((JTable)component).scrollRectToVisible(cellRect);
        }
//        // use it instead
        screenLocation = new Point(cellRect.x,cellRect.y);
        SwingUtilities.convertPointToScreen(screenLocation, insideComponent);
        insideComponentSize = new Dimension(cellRect.width, cellRect.height);
      }
			Point location = new Point();
			Rectangle sBounds = insideComponent.getGraphicsConfiguration()
					.getBounds();
			boolean leftToRight = insideComponent.getComponentOrientation()
					.isLeftToRight();

			// Just to be paranoid
			hideInputTip();

			tip = insideComponent.createToolTip();
			tip.setTipText(tipText);
			size = tip.getPreferredSize();
			int bottomEdgeOfComponent = screenLocation.y
					 + ((int) insideComponentSize.getHeight());
			location.x = screenLocation.x;
//            location.y = screenLocation.y + ((int) insideComponent.getHeight())
//                    + 10;
			location.y = screenLocation.y - 2 - size.height;
			if (!leftToRight) {
				if (location.x - size.width >= 0) {
					location.x -= size.width;
				}
			}
			if (popupRect == null) {
				popupRect = new Rectangle();
			}
			popupRect.setBounds(location.x, location.y, size.width, size.height);

			int y = getPopupFitHeight(popupRect, insideComponent);
			int x = getPopupFitWidth(popupRect, insideComponent);

			if (y > 0) {
				location.y -= y;
			}
			if (x > 0) {
				// adjust
				location.x -= x;
			}

			// Fit as much of the tooltip on screen as possible
			if (location.x < sBounds.x) {
				location.x = sBounds.x;
			} else if (location.x - sBounds.x + size.width > sBounds.width) {
				location.x = sBounds.x
						 + Math.max(0, sBounds.width - size.width);
			}
			if (location.y < sBounds.y) {
				location.y = sBounds.y;
			} else if (location.y - sBounds.y + size.height > sBounds.height) {
				location.y = sBounds.y
						 + Math.max(0, sBounds.height - size.height);
			}

			PopupFactory popupFactory = PopupFactory.getSharedInstance();

			tipWindow = popupFactory.getPopup(insideComponent, tip, location.x,
					location.y);
			//     popupFactory.setPopupType(PopupFactory.LIGHT_WEIGHT_POPUP);

			tipWindow.show();
      if (attachFocusListener) {
        insideComponent.addFocusListener(this);
        focusComponent = insideComponent;
        inputTipsKeyHandler = new InputTipsKeyHandler();
        keyboardFocusManager = KeyboardFocusManager.getCurrentKeyboardFocusManager();
        keyboardFocusManager.addKeyEventDispatcher(inputTipsKeyHandler);
      }
			tipShowing = true;
			if (beepEnabled) {
				Toolkit.getDefaultToolkit().beep();
			}
		}
	}


  
	/**
	 *  Hides the InputTip
	 */
	public synchronized void hideInputTip() {
		if (tipWindow != null) {
			tipWindow.hide();
			tipWindow = null;
			tipShowing = false;
			(tip.getUI()).uninstallUI(tip);
			tip = null;
		}
		if (focusComponent != null) {
			focusComponent.removeFocusListener(this);
		}
		focusComponent = null;// clear it.
		if (inputTipsKeyHandler != null && keyboardFocusManager != null) {
			keyboardFocusManager.removeKeyEventDispatcher(inputTipsKeyHandler);
		}
		keyboardFocusManager = null;
		inputTipsKeyHandler = null;
	}


	/**
	 *  Returns a shared <code>InputTipsManager</code> instance.
	 *
	 *@return    a shared <code>InputTipsManager</code> object
	 */
	public static InputTipsManager sharedInstance() {
		return sharedInstance;
	}


	/**
	 *  Gets a display text for the key in a <code>KeyEvent</code>
	 *
	 *@param  e  the KeyEvent
	 *@return    The display text for the key in this event
	 */
	public static String getKeyText(KeyEvent e) {
		StringBuffer str = new StringBuffer();
		char c = e.getKeyChar();
		switch (c) {
						case KeyEvent.CHAR_UNDEFINED:
							str.append(KeyEvent.getKeyText(e.getKeyCode()));
							break;
						default:
							str.append(getKeyText(c));
							break;
		}
		return str.toString();
	}


	/**
	 *  Gets a display text for this <code>char</code><br>
	 *  Has special handling for Backspace, Tab, Enter, Cancel, Escape, Delete and
	 *  SYN (code 0x16)
	 *
	 *@param  c  the char to display
	 *@return    The display text for this char
	 */
	public static String getKeyText(char c) {
		StringBuffer str = new StringBuffer();
		switch (c) {
						case '\b':
							str.append(getKeyText(KeyEvent.VK_BACK_SPACE));
							break;
						case '\t':
							str.append(getKeyText(KeyEvent.VK_TAB));
							break;
						case '\n':
							str.append(getKeyText(KeyEvent.VK_ENTER));
							break;
						case '\u0018':
							str.append(getKeyText(KeyEvent.VK_CANCEL));
							break;
						case '\u0016':
							str.append("SYN");
							break;
						case '\u001b':
							str.append(getKeyText(KeyEvent.VK_ESCAPE));
							break;
						case '\u007f':
							str.append(getKeyText(KeyEvent.VK_DELETE));
							break;
						case KeyEvent.CHAR_UNDEFINED:
							str.append(Toolkit.getProperty("AWT.undefinedKeyChar",
									"Undefined keyChar"));
							break;
						default:
							str.append("'").append(c).append("'");
							break;
		}
		return str.toString();
	}


	/**
	 *  This class hides the inputtip on each keystroke press
	 *
	 *@author     Matthew Ford
	 *@version    1.3
	 *created    October 19, 2004
	 */
	protected class InputTipsKeyHandler implements KeyEventDispatcher {
		/**
		 *  Description of the Method
		 *
		 *@param  e  Description of the Parameter
		 *@return    Description of the Return Value
		 */
		public boolean dispatchKeyEvent(KeyEvent e) {
			if (e.getID() == KeyEvent.KEY_PRESSED) {
				// hide tip on any keypressed
				InputTipsManager.sharedInstance().hideInputTip();
			}
			return false;
		}
	}


	/**
	 *  Gets a display text for this keycode e.g. "HOME", "F1" or "A". These
	 *  strings can be localized by changing the awt.properties file.
	 *
	 *@param  keyCode  the keycode to translate
	 *@return          The display text for this key code
	 */
	public static String getKeyText(int keyCode) {
		if (keyCode >= KeyEvent.VK_0 && keyCode <= KeyEvent.VK_9
				 || keyCode >= KeyEvent.VK_A && keyCode <= KeyEvent.VK_Z) {
			return String.valueOf((char) keyCode);
		}

		// Check for other ASCII keyCodes.
		int index = ",./;=[\\]".indexOf(keyCode);
		if (index >= 0) {
			return String.valueOf((char) keyCode);
		}

		switch (keyCode) {
						case KeyEvent.VK_ENTER:
							return Toolkit.getProperty("AWT.enter", "Enter");
						case KeyEvent.VK_BACK_SPACE:
							return Toolkit.getProperty("AWT.backSpace", "Backspace");
						case KeyEvent.VK_TAB:
							return Toolkit.getProperty("AWT.tab", "Tab");
						case KeyEvent.VK_CANCEL:
							return Toolkit.getProperty("AWT.cancel", "Cancel");
						case KeyEvent.VK_CLEAR:
							return Toolkit.getProperty("AWT.clear", "Clear");
						case KeyEvent.VK_SHIFT:
							return Toolkit.getProperty("AWT.shift", "Shift");
						case KeyEvent.VK_CONTROL:
							return Toolkit.getProperty("AWT.control", "Control");
						case KeyEvent.VK_ALT:
							return Toolkit.getProperty("AWT.alt", "Alt");
						case KeyEvent.VK_PAUSE:
							return Toolkit.getProperty("AWT.pause", "Pause");
						case KeyEvent.VK_CAPS_LOCK:
							return Toolkit.getProperty("AWT.capsLock", "Caps Lock");
						case KeyEvent.VK_ESCAPE:
							return Toolkit.getProperty("AWT.escape", "Escape");
						case KeyEvent.VK_SPACE:
							return Toolkit.getProperty("AWT.space", "Space");
						case KeyEvent.VK_PAGE_UP:
							return Toolkit.getProperty("AWT.pgup", "Page Up");
						case KeyEvent.VK_PAGE_DOWN:
							return Toolkit.getProperty("AWT.pgdn", "Page Down");
						case KeyEvent.VK_END:
							return Toolkit.getProperty("AWT.end", "End");
						case KeyEvent.VK_HOME:
							return Toolkit.getProperty("AWT.home", "Home");
						case KeyEvent.VK_LEFT:
							return Toolkit.getProperty("AWT.left", "Left");
						case KeyEvent.VK_UP:
							return Toolkit.getProperty("AWT.up", "Up");
						case KeyEvent.VK_RIGHT:
							return Toolkit.getProperty("AWT.right", "Right");
						case KeyEvent.VK_DOWN:
							return Toolkit.getProperty("AWT.down", "Down");
						// handled by ASCII value above:
						//   comma, period, slash, 0-9, semicolon, equals, A-Z,
						//   open_bracket, back_slash, close_bracket

						// numpad numeric keys handled below
						case KeyEvent.VK_MULTIPLY:
							return Toolkit.getProperty("AWT.multiply", "NumPad *");
						case KeyEvent.VK_ADD:
							return Toolkit.getProperty("AWT.add", "NumPad +");
						case KeyEvent.VK_SEPARATOR:
							return Toolkit.getProperty("AWT.separator", "NumPad ,");
						case KeyEvent.VK_SUBTRACT:
							return Toolkit.getProperty("AWT.subtract", "NumPad -");
						case KeyEvent.VK_DECIMAL:
							return Toolkit.getProperty("AWT.decimal", "NumPad .");
						case KeyEvent.VK_DIVIDE:
							return Toolkit.getProperty("AWT.divide", "NumPad /");
						case KeyEvent.VK_DELETE:
							return Toolkit.getProperty("AWT.delete", "Delete");
						case KeyEvent.VK_NUM_LOCK:
							return Toolkit.getProperty("AWT.numLock", "Num Lock");
						case KeyEvent.VK_SCROLL_LOCK:
							return Toolkit.getProperty("AWT.scrollLock", "Scroll Lock");
						case KeyEvent.VK_F1:
							return Toolkit.getProperty("AWT.f1", "F1");
						case KeyEvent.VK_F2:
							return Toolkit.getProperty("AWT.f2", "F2");
						case KeyEvent.VK_F3:
							return Toolkit.getProperty("AWT.f3", "F3");
						case KeyEvent.VK_F4:
							return Toolkit.getProperty("AWT.f4", "F4");
						case KeyEvent.VK_F5:
							return Toolkit.getProperty("AWT.f5", "F5");
						case KeyEvent.VK_F6:
							return Toolkit.getProperty("AWT.f6", "F6");
						case KeyEvent.VK_F7:
							return Toolkit.getProperty("AWT.f7", "F7");
						case KeyEvent.VK_F8:
							return Toolkit.getProperty("AWT.f8", "F8");
						case KeyEvent.VK_F9:
							return Toolkit.getProperty("AWT.f9", "F9");
						case KeyEvent.VK_F10:
							return Toolkit.getProperty("AWT.f10", "F10");
						case KeyEvent.VK_F11:
							return Toolkit.getProperty("AWT.f11", "F11");
						case KeyEvent.VK_F12:
							return Toolkit.getProperty("AWT.f12", "F12");
						case KeyEvent.VK_F13:
							return Toolkit.getProperty("AWT.f13", "F13");
						case KeyEvent.VK_F14:
							return Toolkit.getProperty("AWT.f14", "F14");
						case KeyEvent.VK_F15:
							return Toolkit.getProperty("AWT.f15", "F15");
						case KeyEvent.VK_F16:
							return Toolkit.getProperty("AWT.f16", "F16");
						case KeyEvent.VK_F17:
							return Toolkit.getProperty("AWT.f17", "F17");
						case KeyEvent.VK_F18:
							return Toolkit.getProperty("AWT.f18", "F18");
						case KeyEvent.VK_F19:
							return Toolkit.getProperty("AWT.f19", "F19");
						case KeyEvent.VK_F20:
							return Toolkit.getProperty("AWT.f20", "F20");
						case KeyEvent.VK_F21:
							return Toolkit.getProperty("AWT.f21", "F21");
						case KeyEvent.VK_F22:
							return Toolkit.getProperty("AWT.f22", "F22");
						case KeyEvent.VK_F23:
							return Toolkit.getProperty("AWT.f23", "F23");
						case KeyEvent.VK_F24:
							return Toolkit.getProperty("AWT.f24", "F24");
						case KeyEvent.VK_PRINTSCREEN:
							return Toolkit.getProperty("AWT.printScreen", "Print Screen");
						case KeyEvent.VK_INSERT:
							return Toolkit.getProperty("AWT.insert", "Insert");
						case KeyEvent.VK_HELP:
							return Toolkit.getProperty("AWT.help", "Help");
						case KeyEvent.VK_META:
							return Toolkit.getProperty("AWT.meta", "Meta");
						case KeyEvent.VK_BACK_QUOTE:
							return Toolkit.getProperty("AWT.backQuote", "Back Quote");
						case KeyEvent.VK_QUOTE:
							return Toolkit.getProperty("AWT.quote", "Quote");
						case KeyEvent.VK_KP_UP:
							return Toolkit.getProperty("AWT.up", "Up");
						case KeyEvent.VK_KP_DOWN:
							return Toolkit.getProperty("AWT.down", "Down");
						case KeyEvent.VK_KP_LEFT:
							return Toolkit.getProperty("AWT.left", "Left");
						case KeyEvent.VK_KP_RIGHT:
							return Toolkit.getProperty("AWT.right", "Right");
						case KeyEvent.VK_DEAD_GRAVE:
							return Toolkit.getProperty("AWT.deadGrave", "Dead Grave");
						case KeyEvent.VK_DEAD_ACUTE:
							return Toolkit.getProperty("AWT.deadAcute", "Dead Acute");
						case KeyEvent.VK_DEAD_CIRCUMFLEX:
							return Toolkit.getProperty("AWT.deadCircumflex", "Dead Circumflex");
						case KeyEvent.VK_DEAD_TILDE:
							return Toolkit.getProperty("AWT.deadTilde", "Dead Tilde");
						case KeyEvent.VK_DEAD_MACRON:
							return Toolkit.getProperty("AWT.deadMacron", "Dead Macron");
						case KeyEvent.VK_DEAD_BREVE:
							return Toolkit.getProperty("AWT.deadBreve", "Dead Breve");
						case KeyEvent.VK_DEAD_ABOVEDOT:
							return Toolkit.getProperty("AWT.deadAboveDot", "Dead Above Dot");
						case KeyEvent.VK_DEAD_DIAERESIS:
							return Toolkit.getProperty("AWT.deadDiaeresis", "Dead Diaeresis");
						case KeyEvent.VK_DEAD_ABOVERING:
							return Toolkit.getProperty("AWT.deadAboveRing", "Dead Above Ring");
						case KeyEvent.VK_DEAD_DOUBLEACUTE:
							return Toolkit.getProperty("AWT.deadDoubleAcute",
									"Dead Double Acute");
						case KeyEvent.VK_DEAD_CARON:
							return Toolkit.getProperty("AWT.deadCaron", "Dead Caron");
						case KeyEvent.VK_DEAD_CEDILLA:
							return Toolkit.getProperty("AWT.deadCedilla", "Dead Cedilla");
						case KeyEvent.VK_DEAD_OGONEK:
							return Toolkit.getProperty("AWT.deadOgonek", "Dead Ogonek");
						case KeyEvent.VK_DEAD_IOTA:
							return Toolkit.getProperty("AWT.deadIota", "Dead Iota");
						case KeyEvent.VK_DEAD_VOICED_SOUND:
							return Toolkit.getProperty("AWT.deadVoicedSound",
									"Dead Voiced Sound");
						case KeyEvent.VK_DEAD_SEMIVOICED_SOUND:
							return Toolkit.getProperty("AWT.deadSemivoicedSound",
									"Dead Semivoiced Sound");
						case KeyEvent.VK_AMPERSAND:
							return Toolkit.getProperty("AWT.ampersand", "Ampersand");
						case KeyEvent.VK_ASTERISK:
							return Toolkit.getProperty("AWT.asterisk", "Asterisk");
						case KeyEvent.VK_QUOTEDBL:
							return Toolkit.getProperty("AWT.quoteDbl", "Double Quote");
						case KeyEvent.VK_LESS:
							return Toolkit.getProperty("AWT.Less", "Less");
						case KeyEvent.VK_GREATER:
							return Toolkit.getProperty("AWT.greater", "Greater");
						case KeyEvent.VK_BRACELEFT:
							return Toolkit.getProperty("AWT.braceLeft", "Left Brace");
						case KeyEvent.VK_BRACERIGHT:
							return Toolkit.getProperty("AWT.braceRight", "Right Brace");
						case KeyEvent.VK_AT:
							return Toolkit.getProperty("AWT.at", "At");
						case KeyEvent.VK_COLON:
							return Toolkit.getProperty("AWT.colon", "Colon");
						case KeyEvent.VK_CIRCUMFLEX:
							return Toolkit.getProperty("AWT.circumflex", "Circumflex");
						case KeyEvent.VK_DOLLAR:
							return Toolkit.getProperty("AWT.dollar", "Dollar");
						case KeyEvent.VK_EURO_SIGN:
							return Toolkit.getProperty("AWT.euro", "Euro");
						case KeyEvent.VK_EXCLAMATION_MARK:
							return Toolkit.getProperty("AWT.exclamationMark",
									"Exclamation Mark");
						case KeyEvent.VK_INVERTED_EXCLAMATION_MARK:
							return Toolkit.getProperty("AWT.invertedExclamationMark",
									"Inverted Exclamation Mark");
						case KeyEvent.VK_LEFT_PARENTHESIS:
							return Toolkit.getProperty("AWT.leftParenthesis",
									"Left Parenthesis");
						case KeyEvent.VK_NUMBER_SIGN:
							return Toolkit.getProperty("AWT.numberSign", "Number Sign");
						case KeyEvent.VK_MINUS:
							return Toolkit.getProperty("AWT.minus", "Minus");
						case KeyEvent.VK_PLUS:
							return Toolkit.getProperty("AWT.plus", "Plus");
						case KeyEvent.VK_RIGHT_PARENTHESIS:
							return Toolkit.getProperty("AWT.rightParenthesis",
									"Right Parenthesis");
						case KeyEvent.VK_UNDERSCORE:
							return Toolkit.getProperty("AWT.underscore", "Underscore");
						case KeyEvent.VK_FINAL:
							return Toolkit.getProperty("AWT.final", "Final");
						case KeyEvent.VK_CONVERT:
							return Toolkit.getProperty("AWT.convert", "Convert");
						case KeyEvent.VK_NONCONVERT:
							return Toolkit.getProperty("AWT.noconvert", "No Convert");
						case KeyEvent.VK_ACCEPT:
							return Toolkit.getProperty("AWT.accept", "Accept");
						case KeyEvent.VK_MODECHANGE:
							return Toolkit.getProperty("AWT.modechange", "Mode Change");
						case KeyEvent.VK_KANA:
							return Toolkit.getProperty("AWT.kana", "Kana");
						case KeyEvent.VK_KANJI:
							return Toolkit.getProperty("AWT.kanji", "Kanji");
						case KeyEvent.VK_ALPHANUMERIC:
							return Toolkit.getProperty("AWT.alphanumeric", "Alphanumeric");
						case KeyEvent.VK_KATAKANA:
							return Toolkit.getProperty("AWT.katakana", "Katakana");
						case KeyEvent.VK_HIRAGANA:
							return Toolkit.getProperty("AWT.hiragana", "Hiragana");
						case KeyEvent.VK_FULL_WIDTH:
							return Toolkit.getProperty("AWT.fullWidth", "Full-Width");
						case KeyEvent.VK_HALF_WIDTH:
							return Toolkit.getProperty("AWT.halfWidth", "Half-Width");
						case KeyEvent.VK_ROMAN_CHARACTERS:
							return Toolkit.getProperty("AWT.romanCharacters",
									"Roman Characters");
						case KeyEvent.VK_ALL_CANDIDATES:
							return Toolkit.getProperty("AWT.allCandidates", "All Candidates");
						case KeyEvent.VK_PREVIOUS_CANDIDATE:
							return Toolkit.getProperty("AWT.previousCandidate",
									"Previous Candidate");
						case KeyEvent.VK_CODE_INPUT:
							return Toolkit.getProperty("AWT.codeInput", "Code Input");
						case KeyEvent.VK_JAPANESE_KATAKANA:
							return Toolkit.getProperty("AWT.japaneseKatakana",
									"Japanese Katakana");
						case KeyEvent.VK_JAPANESE_HIRAGANA:
							return Toolkit.getProperty("AWT.japaneseHiragana",
									"Japanese Hiragana");
						case KeyEvent.VK_JAPANESE_ROMAN:
							return Toolkit.getProperty("AWT.japaneseRoman", "Japanese Roman");
						case KeyEvent.VK_KANA_LOCK:
							return Toolkit.getProperty("AWT.kanaLock", "Kana Lock");
						case KeyEvent.VK_INPUT_METHOD_ON_OFF:
							return Toolkit.getProperty("AWT.inputMethodOnOff",
									"Input Method On/Off");
						case KeyEvent.VK_AGAIN:
							return Toolkit.getProperty("AWT.again", "Again");
						case KeyEvent.VK_UNDO:
							return Toolkit.getProperty("AWT.undo", "Undo");
						case KeyEvent.VK_COPY:
							return Toolkit.getProperty("AWT.copy", "Copy");
						case KeyEvent.VK_PASTE:
							return Toolkit.getProperty("AWT.paste", "Paste");
						case KeyEvent.VK_CUT:
							return Toolkit.getProperty("AWT.cut", "Cut");
						case KeyEvent.VK_FIND:
							return Toolkit.getProperty("AWT.find", "Find");
						case KeyEvent.VK_PROPS:
							return Toolkit.getProperty("AWT.props", "Props");
						case KeyEvent.VK_STOP:
							return Toolkit.getProperty("AWT.stop", "Stop");
						case KeyEvent.VK_COMPOSE:
							return Toolkit.getProperty("AWT.compose", "Compose");
						case KeyEvent.VK_ALT_GRAPH:
							return Toolkit.getProperty("AWT.altGraph", "Alt Graph");
		}

		if (keyCode >= KeyEvent.VK_NUMPAD0 && keyCode <= KeyEvent.VK_NUMPAD9) {
			String numpad = Toolkit.getProperty("AWT.numpad", "NumPad");
			char c = (char) (keyCode - KeyEvent.VK_NUMPAD0 + '0');
			return numpad + "-" + c;
		}

		//String unknown = Toolkit.getProperty("AWT.unknown",
		//                   "Unknown keyCode");
		return "" + ((char) keyCode);
		//or ": 0x" + Integer.toString(keyCode, 16);
	}


	/**
	 *  Get the Frame containing this component
	 *
	 *@param  component  the component
	 *@return            the Frame containing it
	 */
	protected static Frame frameForComponent(Component component) {
		while (!(component instanceof Frame)) {
			component = component.getParent();
		}
		return (Frame) component;
	}

	// Returns: 0 no adjust
	//         -1 can't fit
	//         >0 adjust value by amount returned
	/**
	 *  Gets the popupFitWidth attribute of the InputTipsManager object
	 *
	 *@param  popupRectInScreen  Description of the Parameter
	 *@param  invoker            Description of the Parameter
	 *@return                    The popupFitWidth value
	 */
	private int getPopupFitWidth(Rectangle popupRectInScreen, Component invoker) {
		if (invoker != null) {
			Container parent;
			for (parent = invoker.getParent(); parent != null; parent = parent.getParent()) {
				// fix internal frame size bug: 4139087 - 4159012
				if (parent instanceof JFrame || parent instanceof JDialog
						 || parent instanceof JWindow) {// no check for
					// awt.Frame since we
					// use Heavy tips
					return getWidthAdjust(parent.getBounds(), popupRectInScreen);
				} else if (parent instanceof JApplet
						 || parent instanceof JInternalFrame) {
					if (popupFrameRect == null) {
						popupFrameRect = new Rectangle();
					}
					Point p = parent.getLocationOnScreen();
					popupFrameRect.setBounds(p.x, p.y, parent.getBounds().width,
							parent.getBounds().height);
					return getWidthAdjust(popupFrameRect, popupRectInScreen);
				}
			}
		}
		return 0;
	}

	// Returns: 0 no adjust
	//          >0 adjust by value return
	/**
	 *  Gets the popupFitHeight attribute of the InputTipsManager object
	 *
	 *@param  popupRectInScreen  Description of the Parameter
	 *@param  invoker            Description of the Parameter
	 *@return                    The popupFitHeight value
	 */
	private int getPopupFitHeight(Rectangle popupRectInScreen, Component invoker) {
		if (invoker != null) {
			Container parent;
			for (parent = invoker.getParent(); parent != null; parent = parent.getParent()) {
				if (parent instanceof JFrame || parent instanceof JDialog
						 || parent instanceof JWindow) {
					return getHeightAdjust(parent.getBounds(),
							popupRectInScreen);
				} else if (parent instanceof JApplet
						 || parent instanceof JInternalFrame) {
					if (popupFrameRect == null) {
						popupFrameRect = new Rectangle();
					}
					Point p = parent.getLocationOnScreen();
					popupFrameRect.setBounds(p.x, p.y, parent.getBounds().width,
							parent.getBounds().height);
					return getHeightAdjust(popupFrameRect, popupRectInScreen);
				}
			}
		}
		return 0;
	}


	/**
	 *  Gets the heightAdjust attribute of the InputTipsManager object
	 *
	 *@param  a  Description of the Parameter
	 *@param  b  Description of the Parameter
	 *@return    The heightAdjust value
	 */
	private int getHeightAdjust(Rectangle a, Rectangle b) {
		if (b.y >= a.y && (b.y + b.height) <= (a.y + a.height)) {
			return 0;
		} else {
			return (((b.y + b.height) - (a.y + a.height)) + 5);
		}
	}

	// Return the number of pixels over the edge we are extending.
	// If we are over the edge the ToolTipManager can adjust.
	// REMIND: what if the Tooltip is just too big to fit at all - we currently
	// will just clip
	/**
	 *  Gets the widthAdjust attribute of the InputTipsManager object
	 *
	 *@param  a  Description of the Parameter
	 *@param  b  Description of the Parameter
	 *@return    The widthAdjust value
	 */
	private int getWidthAdjust(Rectangle a, Rectangle b) {
		//    System.out.println("width b.x/b.width: " + b.x + "/" + b.width +
		//		       "a.x/a.width: " + a.x + "/" + a.width);
		if (b.x >= a.x && (b.x + b.width) <= (a.x + a.width)) {
			return 0;
		} else {
			return (((b.x + b.width) - (a.x + a.width)) + 5);
		}
	}


}

